//! Dispatching to state service task(s) through a supervisor
use super::*;
use tokio::time;
use std::{fmt, error};
use futures::prelude::*;

impl Supervisor
{
    /// Dispatch a request to the supervisor to be passed through to a subtask.
    ///
    /// # Returns
    /// Returns a `Future` that can be awaited on to produce the value sent back by the task (if there is one).
    ///
    /// # Errors
    /// * The first failure will be caused if sending to the supervisor fails.
    /// * The 2nd failure will be caused if either the supervisor, or its delegated task panics before being able to respond, or if the task simply does not respond.
    pub async fn dispatch_req(&mut self, kind: ServiceRequestKind) -> Result<impl Future<Output=Result<ServiceResponse, SupervisorDispatchError>> + 'static, SupervisorDispatchError>
    {
	let (req, rx) = ServiceRequest::new(kind);
	self.pipe.send(req).await.map_err(|_| SupervisorDispatchError::Send)?;
	Ok(rx.map_err(|_| SupervisorDispatchError::Recv))
    }

    /// Dispatch a request to the supervisor to be passed through to a subtask and then wait for a response from it.
    ///
    /// # Returns
    /// Returns the value sent back by the task, if there is one
    pub async fn dispatch_and_wait(&mut self, kind: ServiceRequestKind) -> Result<ServiceResponse, SupervisorDispatchError>
    {
	Ok(self.dispatch_req(kind)
	   .await?
	   .await?)
    }
    
    /// Dispatch a request to the supervisor to be passed through to a subtask and then wait for a response from it.
    /// If the timeout expires before a response from the server is received, then the operation will cancel and the error returned will be `SupervisorDispatchError::Timeout`. 
    ///
    /// # Returns
    /// Returns the value sent back by the task, if there is one
    pub async fn dispatch_and_wait_timeout(&mut self, kind: ServiceRequestKind, timeout: time::Duration) -> Result<ServiceResponse, SupervisorDispatchError>
    {
	let resp_wait = self.dispatch_req(kind)
	    .await?;
	tokio::select! {
	    val = resp_wait => {
		return Ok(val?);
	    }
	    _ = time::delay_for(timeout) => {
		return Err(SupervisorDispatchError::Timeout("receiving response", Some(timeout)))
	    }
	}
    }    
    /// Dispatch a request to the supervisor to be passed through to a subtask and then wait for a response from it.
    /// If the future `until` completes before a response from the server is received, then the operation will cancel and the error returned will be `SupervisorDispatchError::Timeout`. 
    ///
    /// # Returns
    /// Returns the value sent back by the task, if there is one
    pub async fn dispatch_and_wait_until(&mut self, kind: ServiceRequestKind, until: impl Future) -> Result<ServiceResponse, SupervisorDispatchError>
    {
	let resp_wait = self.dispatch_req(kind)
	    .await?;
	tokio::select! {
	    val = resp_wait => {
		return Ok(val?);
	    }
	    _ = until => {
		return Err(SupervisorDispatchError::Timeout("receiving response", None))
	    }
	}
    }
}

/// Error when dispatching a request to the supervisor
#[derive(Debug)]
pub enum SupervisorDispatchError
{
    Send, Recv, Timeout(&'static str, Option<tokio::time::Duration>),
}

impl error::Error for SupervisorDispatchError{}
impl fmt::Display for SupervisorDispatchError
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	match self {
	    Self::Send => write!(f, "dispatching the request failed"),
	    Self::Recv => write!(f, "receiving the response failed"),
	    Self::Timeout(msg, Some(duration)) => write!(f, "timeout on {} was reached ({:?})", msg, duration),
	    Self::Timeout(msg, _) => write!(f, "timeout on {} was reached", msg),
	}
    }
}
